import numpy as np
import matplotlib.pyplot as plt
from collections import Counter

def find_breakpoints(x,y):
    """
    Finds the breakpoints of a piecewise linear function.

    Parameters:
        points (list of tuples): List of (x, y) tuples representing the function.

    Returns:
        list: List of x-coordinates that are breakpoints of the function.
    """
    if len(x) < 2:
        raise ValueError("At least two points are needed to define a piecewise linear function.")

    # Sort the points by x-coordinate (if not already sorted)
    # points = sorted(points, key=lambda p: p[0])
    breakpoints = []
    #breakpoints = [x[0]]  # The first x-coordinate is always a breakpoint
    for i in range(1, len(x) - 1):
        x1, y1 = x[i - 1], y[i-1]
        x2, y2 = x[i], y[i]
        x3, y3 = x[i + 1], y[i+1]

        # Calculate slopes of adjacent segments
        slope1 = (y2 - y1) / (x2 - x1)
        slope2 = (y3 - y2) / (x3 - x2)

        # Check if there's a change in slope
        if abs(slope1 - slope2) >= 1:
            breakpoints.append(x2)
            

    #breakpoints.append(x[-1])  # The last x-coordinate is always a breakpoint

    return breakpoints



def sigma(x):
    return np.minimum(np.maximum(0, x),1)



N_MC = 1000
bkp = [0] * N_MC
random_width = 6
weight_mag = 4

draw_pts = 2**5 +1
weights_history = [None]* N_MC

for i in range(N_MC):
    W1 = np.random.randint(-1*weight_mag, weight_mag, size=(random_width, 1))
    W2 = np.random.randint(-1*weight_mag, weight_mag, size=(random_width, random_width))
    W3 = np.random.randint(-1*weight_mag, weight_mag, size=(random_width, random_width))
    W4 = np.random.randint(-1*weight_mag, weight_mag, size=(1, random_width))
    # weights_history[i] = [W1, W3]
    # W1 = np.array([[10]])
    def network(x):
        h1 = sigma(np.dot(W1, x))
        h2 = sigma(np.dot(W2, h1))
        h3 = sigma(np.dot(W3, h2))
        h4 = sigma(np.dot(W4, h3))
        return h4
    
    x = np.array(np.linspace(0, 1, draw_pts)).reshape((1,draw_pts))  # 500 points for a smooth curve
    y = network(x)
    bkp[i] = len(find_breakpoints(x.reshape((draw_pts,)),y.reshape((draw_pts,))))
    # print(bkp[i])


# a random 



counts = Counter(bkp)

# Get unique elements and their counts
sorted_unique_elements = sorted(counts.keys())
sorted_occurrences = [counts[element] for element in sorted_unique_elements]



plt.figure(figsize=(8, 6))
plt.plot(x.reshape((draw_pts,)), y.reshape((draw_pts,)), color="blue")
plt.xlabel("x", fontsize=12)
plt.ylabel("y", fontsize=12)
plt.grid(True, linestyle="--", alpha=0.6)
plt.legend(fontsize=12)
plt.show()


# Example usage
piecewise_points = [(0, 0), (1, 2), (2, 4), (3, 3), (4, 1), (5, 1)]
breakpoints = find_breakpoints(piecewise_points)
print("Breakpoints:", breakpoints)

