import numpy as np
from config import Symbols
from valuation import find_outer_connective
from matplotlib import pyplot as plt
from activations import relu
from generate_frac import frac_div, frac_minus

def check_sign(x):
    if x > 0:
        return 1
    if x < 0:
        return -1
    return 0
def construct_shallow_NN_from_MV(xcords,ycords): ## xcords and ycords are given in the form of fractions
    slopes = [None] * (len(xcords)-1)
    for i in range(len(xcords)-1):
        x_next = xcords[i+1]
        y_next = ycords[i+1]
        x = xcords[i]
        y = ycords[i]
        slope = frac_div(frac_minus(y_next, y), frac_minus(x_next, x))
        slopes[i] = slope[0]//slope[1]
    slopes.append(0)
    slopes.insert(0,0)
    slopes_diff = [slopes[i+1]-slopes[i] for i in range(len(slopes)-1)]
    # slopes_diff = slopes_diff.pop() if slopes_diff[-1] == 0 else slopes_diff
    #if slopes[0] != 0:
    slopes_diff_abs = [abs(i) for i in slopes_diff]
    W1 = np.array([slopes_diff_abs]).reshape((-1,1))
    b1 = [-1*abs(slopes_diff[i] * xcords[i][0])//xcords[i][1] for i in range(len(xcords))]
    b1 = np.array([b1]).reshape((-1,1))
    W2 = [check_sign(i) for i in slopes_diff]
    W2 = np.array([W2]).reshape((1,-1))
    b2 = ycords[0][0]//ycords[0][1]
    b2 = np.array([[b2]])
    Phi = [W1, b1, W2, b2]
    return Phi

def construct_NN_from_MV_multi_3(mv):
    Phi = []
    if mv=='x1':
        Phi.append(np.array([[1, 0, 0]]))
        Phi.append(np.array([[0]]))
        Phi.append(np.array([[1]]))
        Phi.append(np.array([[0]]))
        return Phi
    if mv ==f"({Symbols.NOT}x1)":
        Phi.append(np.array([[-1, 0, 0]]))
        Phi.append(np.array([[1]]))
        Phi.append(np.array([[1]]))
        Phi.append(np.array([[0]]))
        return Phi
    if mv=='x2':
        Phi.append(np.array([[0, 1, 0]]))
        Phi.append(np.array([[0]]))
        Phi.append(np.array([[1]]))
        Phi.append(np.array([[0]]))
        return Phi
    if mv ==f"({Symbols.NOT}x2)":  
        Phi.append(np.array([[0, -1, 0]]))
        Phi.append(np.array([[1]]))
        Phi.append(np.array([[1]]))
        Phi.append(np.array([[0]]))
        return Phi
    if mv=='x3':
        Phi.append(np.array([[0, 0, 1]]))
        Phi.append(np.array([[0]]))
        Phi.append(np.array([[1]]))
        Phi.append(np.array([[0]]))
        return Phi
    if mv ==f"({Symbols.NOT}x3)":
        Phi.append(np.array([[0, 0, -1]]))
        Phi.append(np.array([[1]]))
        Phi.append(np.array([[1]]))
        Phi.append(np.array([[0]]))
        return Phi

    if mv[1] == Symbols.NOT:
        Phi_not_mv = construct_NN_from_MV_multi_3(mv[2:-1])
        W = Phi_not_mv[-2]
        b = Phi_not_mv[-1]
        W = -W
        b = -b+1
        Phi_not_mv[-2] = W
        Phi_not_mv[-1] = b
        return Phi_not_mv

    outer_connective_index = find_outer_connective(mv)
    left, outer_connective, right =mv[1:outer_connective_index], mv[outer_connective_index], mv[outer_connective_index+1:-1]
    left_Phi = construct_NN_from_MV_multi_3(left)
    right_Phi = construct_NN_from_MV_multi_3(right)
    Phi = nn_parallel(left_Phi, right_Phi)
    if outer_connective == Symbols.AND:
        W1 = np.array([[1, 1]])
        Phi.append(W1)
        b1 = np.array([[-1]])
        Phi.append(b1)
        W2 = np.array([[1]])
        Phi.append(W2)
        b2 = np.array([[0]])
        Phi.append(b2)
        return Phi
    if outer_connective == Symbols.OR:
        W1 = np.array([[-1, -1]])
        Phi.append(W1)
        b1 = np.array([[1]])
        Phi.append(b1)
        W2 = np.array([[-1]])
        Phi.append(W2)
        b2 = np.array([[1]])
        Phi.append(b2)
        return Phi
    if outer_connective == Symbols.WEDGE:
        W1 = np.array([[1,0],
                       [-1,0],
                       [1,-1]])
        b1 = np.array([[0],
                       [0],
                       [0]])
        W2 = np.array([[1,-1,-1]])
        b2 = np.array([[0]])
        Phi.append(W1)
        Phi.append(b1)
        Phi.append(W2)
        Phi.append(b2)
        return Phi
    if outer_connective == Symbols.VEE:
        W1 = np.array([[1,0],
                       [-1,0],
                       [-1,1]])
        b1 = np.array([[0],
                       [0],
                       [0]])
        W2 = np.array([[1, -1, 1]])
        b2 = np.array([[0]])
        Phi.append(W1)
        Phi.append(b1)
        Phi.append(W2)
        Phi.append(b2)
        return Phi

def construct_NN_from_MV(mv):
    Phi = []
    if mv == 'x':
        Phi.append(np.array([[1]]))
        Phi.append(np.array([[0]]))
        Phi.append(np.array([[1]]))
        Phi.append(np.array([[0]]))
        return Phi
    if mv == f"({Symbols.NOT}x)":
        Phi.append(np.array([[-1]]))
        Phi.append(np.array([[1]]))
        Phi.append(np.array([[1]]))
        Phi.append(np.array([[0]]))
        return Phi
    if mv[1] == Symbols.NOT:
        Phi_not_mv = construct_NN_from_MV(mv[2:-1])
        W = Phi_not_mv[-2]
        b = Phi_not_mv[-1]
        W = -W
        b = -b+1
        Phi_not_mv[-2] = W
        Phi_not_mv[-1] = b
        return Phi_not_mv

    outer_connective_index = find_outer_connective(mv)
    left, outer_connective, right =mv[1:outer_connective_index], mv[outer_connective_index], mv[outer_connective_index+1:-1]
    left_Phi = construct_NN_from_MV(left)
    right_Phi = construct_NN_from_MV(right)
    Phi = nn_parallel(left_Phi, right_Phi)
    if outer_connective == Symbols.AND:
        W1 = np.array([[1, 1]])
        Phi.append(W1)
        b1 = np.array([[-1]])
        Phi.append(b1)
        W2 = np.array([[1]])
        Phi.append(W2)
        b2 = np.array([[0]])
        Phi.append(b2)
        return Phi
    if outer_connective == Symbols.OR:
        W1 = np.array([[-1, -1]])
        Phi.append(W1)
        b1 = np.array([[1]])
        Phi.append(b1)
        W2 = np.array([[-1]])
        Phi.append(W2)
        b2 = np.array([[1]])
        Phi.append(b2)
        return Phi
    if outer_connective == Symbols.WEDGE:
        W1 = np.array([[1,0],
                       [-1,0],
                       [1,-1]])
        b1 = np.array([[0],
                       [0],
                       [0]])
        W2 = np.array([[1,-1,-1]])
        b2 = np.array([[0]])
        Phi.append(W1)
        Phi.append(b1)
        Phi.append(W2)
        Phi.append(b2)
        return Phi
    if outer_connective == Symbols.VEE:
        W1 = np.array([[1,0],
                       [-1,0],
                       [-1,1]])
        b1 = np.array([[0],
                       [0],
                       [0]])
        W2 = np.array([[1, -1, 1]])
        b2 = np.array([[0]])
        Phi.append(W1)
        Phi.append(b1)
        Phi.append(W2)
        Phi.append(b2)
        return Phi

    


def nn_parallel(Phi_1, Phi_2):
    if len(Phi_1) < len(Phi_2):
        for _ in range((len(Phi_2) - len(Phi_1))//2):
            Phi_1.append(np.array([[1]]))
            Phi_1.append(np.array([[0]]))
    elif len(Phi_1) > len(Phi_2):
        for _ in range((len(Phi_1) - len(Phi_2))//2):
            Phi_2.append(np.array([[1]]))
            Phi_2.append(np.array([[0]]))
    Phi = []
    if len(Phi_1) ==2:
        new_W = np.concatenate((Phi_1[0], Phi_2[0]), axis=0)
        new_b = np.concatenate((Phi_1[1], Phi_2[1]), axis=0)
        Phi.append(new_W)
        Phi.append(new_b)
        return Phi
    new_W = np.concatenate((Phi_1[0], Phi_2[0]), axis=0)
    new_b = np.concatenate((Phi_1[1], Phi_2[1]), axis=0)
    Phi.append(new_W)
    Phi.append(new_b)
    for l in range(2,len(Phi_1)):
        if l % 2 == 0:
            Z1 = np.zeros((Phi_1[l].shape[0],Phi_2[l].shape[1]),dtype=int)
            Z2 = np.zeros((Phi_2[l].shape[0],Phi_1[l].shape[1]),dtype=int)
            new_W = np.asarray(np.bmat([[Phi_1[l], Z1], [Z2, Phi_2[l]]]))
            Phi.append(new_W)
        else:
            new_b = np.concatenate((Phi_1[l], Phi_2[l]), axis=0)
            Phi.append(new_b)
    return Phi


def plot_NN(Phi, denom):
    x = np.linspace(0, 1, denom+1).reshape((1,-1))
    for l in range(len(Phi)//2):
        W = Phi[2*l]
        b = Phi[2*l+1]
        y = relu(np.dot(W, x) + b)
        x = y
    plt.plot(np.linspace(0,1,denom+1), y.reshape((denom+1,)))






    
   