from config import Symbols
import numpy as np

def find_outer_connective(formula):
    if formula[1] == Symbols.NOT: # if the formula is a single variable
        return find_outer_connective(formula[2:-1])+2
    left_bracket = 0
    right_bracket = 0
    for i, c in enumerate(formula):
        if c == '(':
            left_bracket += 1
        if c == ')':
            right_bracket += 1
        if c == Symbols.AND or c == Symbols.OR or c == Symbols.VEE or c == Symbols.WEDGE:
            if left_bracket - right_bracket == 1 and i >= 1:
                return i

# the valuation function, input is a formula and the value of x, output the function value of the formula
def valuation(formula, x):
    if formula == 'x':
        return x
    if formula == f"{Symbols.NOT}x":
        return  1-x
    if formula[0] == Symbols.NOT:
        return  1-valuation(formula[1:], x)
    # find the outer_connective: the first index that left bracket outnumbers the right by 1
    outer_connective_index = find_outer_connective(formula)
    left, outer_connective, right =formula[1:outer_connective_index], formula[outer_connective_index], formula[outer_connective_index+1:-1]
    left_value = valuation(left, x)
    right_value = valuation(right, x)
    if outer_connective == Symbols.AND:
        return np.maximum(left_value + right_value-1,0)
    if outer_connective == Symbols.OR:
        return np.minimum(left_value + right_value,1)


# the fraction version of the valuation function, returned is the numerator of the function value
def valuation_precise(formula, x_num, denom): # x_num is a 1-D array, denom is 1-D array
    if formula == '0':
        return np.array([0]*len(x_num))
    if formula == f"({Symbols.NOT}0)" or formula == f"{Symbols.NOT}0" or formula == '1' or formula == '(1)':
        return denom
        #return np.array([1]*len(x_num))

    if formula == 'x':
        return x_num
    if formula == f"({Symbols.NOT}x)" or formula == f"{Symbols.NOT}x":
        return denom - x_num
    if formula[1] == Symbols.NOT:
        res = valuation_precise(formula[2:-1], x_num, denom)
        return denom - res
        #return  denom-valuation_precise(formula[2:-1], x_num, denom)
    # find the outer_connective: the first index that left bracket outnumbers the right by 1
    outer_connective_index = find_outer_connective(formula)
    left, outer_connective, right =formula[1:outer_connective_index], formula[outer_connective_index], formula[outer_connective_index+1:-1]
    left_value_num = valuation_precise(left, x_num, denom)
    right_value_num = valuation_precise(right, x_num, denom)
    if outer_connective == Symbols.AND:
        return np.maximum(left_value_num + right_value_num-denom,0)
    if outer_connective == Symbols.OR:
        return np.minimum(left_value_num + right_value_num,denom)
    if outer_connective == Symbols.WEDGE:
        return np.minimum(left_value_num, right_value_num)
    if outer_connective == Symbols.VEE:
        return np.maximum(left_value_num, right_value_num)

